#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/time.h>

#include <string>

#include "functions.h"
#include "parser.h"
#include "packet.h"
#include "client.h"
#include "error.h"

SNL_READ_CALLBACK(Client::read_callback) {
	string ret, fnc, par;
	void (*eh)(int err);
	Client *client;
	Packet pkt;

	client = (Client *)data;
	client->mutex.lock();
	eh = client->error_handler;
	if ((error == SNL_ERR_CLOSED) || (error == SNL_ERR_READ)) {
		if (eh) eh(error);
	} else {
		if (pkt.build(buf, len) && (eh)) eh(ODAI_ERR_SYNTAX);
		if (pkt.type() == 's') pkt.pop(ret); else if (eh) eh(ODAI_ERR_SYNTAX);
		if (pkt.type() == 's') pkt.pop(fnc); else if (eh) eh(ODAI_ERR_SYNTAX);
		if (pkt.type() == 's') pkt.pop(par); else if (eh) eh(ODAI_ERR_SYNTAX);
		if (fnc == "return") {
			// store return parameter
			client->rcvd.grab();
			switch (pkt.type()) {
				case 'i': pkt.pop(client->ret_i);     break;
				case 'f': pkt.pop(client->ret_f);     break;
				case 's': pkt.pop(client->ret_s);     break;
				case 'v': /* ignore void parameter */ break;
				case '0': /* ignore end of list */    break;
				default: if (eh) eh(ODAI_ERR_SYNTAX); break;
			}
			client->rcvd.emit();
			client->rcvd.release();
		} else {
			// received a real message
			client->messages.call(fd, buf, len, false); // don't send "return"
		}
	}
	client->mutex.unlock();
}

Client::Client(const char *host, unsigned short port) {
	error_handler = NULL;
	if ((server = snl_connect(host, port)) > 0) {
		snl_read(server, read_callback, this);
	} else {
		throw (ODAI_ERR_SOCKET);
	}
}

Client::~Client() {
	close(server);
}

int
Client::err(void (*fn)(int err)) {
	error_handler = fn;
	return (ODAI_OK);
}

int
Client::reg(const char *fmt, void *fn) {
	string str = "v ";

	str += fmt;
	return (messages.reg(str.c_str(), fn));
}

int
Client::fnc(const char *fmt, ...) {
	va_list args;
	int err;

	va_start(args, fmt);
	err = call('v', NULL, fmt, args);
	va_end(args);
	return (err);
}

int
Client::fnc(int *ret, const char *fmt, ...) {
	va_list args;
	int err;

	va_start(args, fmt);
	err = call('i', (void *)ret, fmt, args);
	va_end(args);
	return (err);
}

int
Client::fnc(double *ret, const char *fmt, ...) {
	va_list args;
	int err;

	va_start(args, fmt);
	err = call('f', (void *)ret, fmt, args);
	va_end(args);
	return (err);
}

int
Client::fnc(char **ret, const char *fmt, ...) {
	va_list args;
	int err;

	va_start(args, fmt);
	err = call('s', (void *)ret, fmt, args);
	va_end(args);
	return (err);
}

int
Client::call(char type, void *ptr, const char *fmt, va_list args) {
	string str, ret, fnc, par;
	Packet pkt;
	int err;

	switch (type) {
		case 'v': str = "v "; str += fmt; break;
		case 'i': str = "i "; str += fmt; break;
		case 'f': str = "f "; str += fmt; break;
		case 's': str = "s "; str += fmt; break;
		default: return (ODAI_ERR_SYNTAX);
	}
	err = pkt.build(str.c_str(), args);
	if (err) return (ODAI_ERR_SYNTAX);
	rcvd.grab();
	mutex.lock();
	err = snl_write(server, pkt.data(), pkt.size());
	mutex.unlock();
	if (err) return (ODAI_ERR_SOCKET);
	// wait for returned telegram (signal)
	if (rcvd.wait()) {
		rcvd.release();
		return (ODAI_ERR_TIMEOUT);
	}
	// received return telegram
	rcvd.release();
	// read return parameter, if caller is intrested in
	if (ptr) {
		switch (type) {
			case 'i': *(int *)    ptr = ret_i;                 break;
			case 'f': *(double *) ptr = ret_f;                 break;
			case 's': *(char **)  ptr = (char *)ret_s.c_str(); break;
			default: return (ODAI_ERR_SYNTAX);
		}
	}
	return (ODAI_OK);
}
